@@ -70,6 +70,7 @@ const char* XPCJSRuntime::mStrings[] = {
     "constructor",          // IDX_CONSTRUCTOR
     "toString",             // IDX_TO_STRING
     "toSource",             // IDX_TO_SOURCE
+    "valueOf",              // IDX_VALUE_OF
     "lastResult",           // IDX_LAST_RESULT
     "returnCode",           // IDX_RETURN_CODE
     "value",                // IDX_VALUE
@@ -127,6 +128,9 @@ struct CX_AND_XPCRT_Data
     XPCJSRuntime* rt;
 };
 
+static void * const UNMARK_ONLY = nsnull;
+static void * const UNMARK_AND_SWEEP = (void *)1;
+
 static JSDHashOperator
 NativeInterfaceSweeper(JSDHashTable *table, JSDHashEntryHdr *hdr,
                        uint32 number, void *arg)
@@ -137,6 +141,9 @@ NativeInterfaceSweeper(JSDHashTable *table, JSDHashEntryHdr *hdr,
         return JS_DHASH_NEXT;
     }
 
+    if (arg == UNMARK_ONLY)
+        return JS_DHASH_NEXT;
+
 #ifdef XPC_REPORT_NATIVE_INTERFACE_AND_SET_FLUSHING
     fputs("- Destroying XPCNativeInterface for ", stdout);
     JS_PutString(JSVAL_TO_STRING(iface->GetName()), stdout);
@@ -172,6 +179,9 @@ NativeSetSweeper(JSDHashTable *table, JSDHashEntryHdr *hdr,
         return JS_DHASH_NEXT;
     }
 
+    if (arg == UNMARK_ONLY)
+        return JS_DHASH_NEXT;
+
 #ifdef XPC_REPORT_NATIVE_INTERFACE_AND_SET_FLUSHING
     printf("- Destroying XPCNativeSet for:\n");
     PRUint16 count = set->GetInterfaceCount();
@@ -203,6 +213,9 @@ JSClassSweeper(JSDHashTable *table, JSDHashEntryHdr *hdr,
         return JS_DHASH_NEXT;
     }
 
+    if (arg == UNMARK_ONLY)
+        return JS_DHASH_NEXT;
+
 #ifdef XPC_REPORT_JSCLASS_FLUSHING
     printf("- Destroying XPCNativeScriptableShared for: %s @ %x\n",
            shared->GetJSClass()->name,
@@ -650,7 +663,7 @@ SweepCompartment(nsCStringHashKey& aKey, JSCompartment *compartment, void *aClos
 }
 
 // static
-JSBool XPCJSRuntime::GCCallback(JSContext *cx, JSGCStatus status)
+JSBool XPCJSRuntime::GCCallback(JSContext *cx, JSGCStatus status, JSBool isCompartmentGC)
 {
     XPCJSRuntime* self = nsXPConnect::GetRuntimeInstance();
     if (!self)
@@ -781,24 +794,37 @@ JSBool XPCJSRuntime::GCCallback(JSContext *cx, JSGCStatus status)
                 }
             }
 
-            // Do the sweeping...
+            // Do the sweeping. During a compartment GC, only
+            // WrappedNativeProtos in collected compartments will be
+            // marked. Therefore, some reachable NativeInterfaces will not be
+            // marked, so it is not safe to sweep them. We still need to unmark
+            // them, since the ones pointed to by WrappedNativeProtos in a
+            // compartment being collected will be marked.
+            //
+            // Ideally, if NativeInterfaces from different compartments were
+            // kept separate, we could sweep only the ones belonging to
+            // compartments being collected. Currently, though, NativeInterfaces
+            // are shared between compartments. This ought to be fixed.
+            void *sweepArg = isCompartmentGC ? UNMARK_ONLY : UNMARK_AND_SWEEP;
 
             // We don't want to sweep the JSClasses at shutdown time.
             // At this point there may be JSObjects using them that have
             // been removed from the other maps.
             if (!self->GetXPConnect()->IsShuttingDown()) {
                 self->mNativeScriptableSharedMap->
-                    Enumerate(JSClassSweeper, nsnull);
+                    Enumerate(JSClassSweeper, sweepArg);
             }
 
-            self->mClassInfo2NativeSetMap->
-                Enumerate(NativeUnMarkedSetRemover, nsnull);
+            if (!isCompartmentGC) {
+                self->mClassInfo2NativeSetMap->
+                    Enumerate(NativeUnMarkedSetRemover, nsnull);
+            }
 
             self->mNativeSetMap->
-                Enumerate(NativeSetSweeper, nsnull);
+                Enumerate(NativeSetSweeper, sweepArg);
 
             self->mIID2NativeInterfaceMap->
-                Enumerate(NativeInterfaceSweeper, nsnull);
+                Enumerate(NativeInterfaceSweeper, sweepArg);
 
 #ifdef DEBUG
             XPCWrappedNativeScope::ASSERT_NoInterfaceSetsAreMarked();
@@ -920,7 +946,7 @@ JSBool XPCJSRuntime::GCCallback(JSContext *cx, JSGCStatus status)
 
     nsTArray<JSGCCallback> callbacks(self->extraGCCallbacks);
     for (PRUint32 i = 0; i < callbacks.Length(); ++i) {
-        if (!callbacks[i](cx, status))
+        if (!callbacks[i](cx, status, isCompartmentGC))
             return JS_FALSE;
     }
 
@@ -2178,6 +2204,7 @@ XPCJSRuntime::XPCJSRuntime(nsXPConnect* aXPConnect)
         JS_SetGrayGCRootsTracer(mJSRuntime, TraceGrayJS, this);
         JS_SetWrapObjectCallbacks(mJSRuntime,
                                   xpc::WrapperFactory::Rewrap,
+                                  xpc::WrapperFactory::WrapForSameCompartment,
                                   xpc::WrapperFactory::PrepareForWrapping);
 #ifdef MOZ_CRASHREPORTER
         JS_EnumerateDiagnosticMemoryRegions(DiagnosticMemoryCallback);
@@ -2296,7 +2323,7 @@ XPCJSRuntime::OnJSContextNew(JSContext *cx)
     return JS_TRUE;
 }
 
-JSBool
+bool
 XPCJSRuntime::DeferredRelease(nsISupports* obj)
 {
     NS_ASSERTION(obj, "bad param");
